/* bttv.c - functions for handling v4l and lowlevel bttv input
 *
 * Parts Copyright (C) Ewald Snel 1998-1999
 * Tuner/Channel stuff from MPlayer v4l interface (C) Alex Beregszaszi
 * BTTV register/reset, disable automute, correct setting of format,
 * Fedora Core 2 fix (C) RB 2004
 * The rest: Copyright (C) 2003 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: bttv.c 59 2004-07-30 21:47:15Z rsnel $
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/io.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/videodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "cc2y4m.h"
#include "bttv.h" 

/* In Fedora Core 2 videodev.h  does not include videodev2.h. We
   need this to turn automute off (see below)
*/
#ifndef __LINUX_VIDEODEV2_H
#define V4L2_CID_PRIVATE_BASE           0x08000000
struct v4l2_control {
	__u32 id;
	__s32 value;
};
#define VIDIOC_G_CTRL           _IOWR ('V', 27, struct v4l2_control)
#define VIDIOC_S_CTRL           _IOWR ('V', 28, struct v4l2_control)
#endif
#define V4L2_CID_PRIVATE_AUTOMUTE    (V4L2_CID_PRIVATE_BASE + 2)

/* Assuming that the capture buffer size is 2129920:
 *
 * A Raw8X frame in TEST mode not all pixels are used; each capture buffer
 * contains: (this data probably depends of the doings of the *Scaler())
 * - 655360 (=10*2^16) pixels of almost continuesly sampled data (288 lines of
 *   2268 pixels and one of 2176 pixels) This data should be treated as the
 *   first field of the captured frame.
 * - 409600 (=100*2^12) unused pixels
 * - 655360 (=10*2^16) pixels of almost continuously sampled data This data
 *   should be treated as the second field 
 * - 409600 (=100*2^12) unused pixels
 *
 * Counting from zero, the pixels 2014, 2015, ..., 2100 are not sampled
 * correctly. the are all too dark, and pixels 2045-2070 are all equal to pixel
 * 2044. This is probalby due to a periodic (every line) reinitiazation of the
 * ADC.
 * 
 * The appearance of the reset with an interval of 2268 pixels tells us that
 * the internal line length really is 2268, otherwise the line length could
 * just as easily be 2270 or 2265 (by a change in the sampling freqency of the
 * ADC). The periodic reset forces us to use lines with a length of 2268
 * pixels. (otherwise we would those reset-artefacts in our active picture or
 * color burst)
 */

/* code from ccd.c, distributed with cabletv-1.3.9,
 * (probably) Copyright (C) Ewald Snel (1998-1999) */
static void doScale(bttv_t *bttv)
{
	local_regs_t *r = bttv->regs;
	short s;
	int field_width = (r->o_hactive_msb << 8) | r->o_hactive_lo;
	int vactive = (r->o_vactive_msb << 8) | r->o_vactive_lo;

	if (bttv->hactive == 0)
		s = 0;
	else
		s =  (short) ((2L*bttv->hdelay*field_width / 
					bttv->hactive + 1) >> 1);

	VERBOSE("doScale, hdelay=%d\n", s);
	r->e_hdelay_msb = r->o_hdelay_msb = (s >> 8) & 3;
	r->e_hdelay_lo = r->o_hdelay_lo = s & 0xff;

	if (field_width == 0)
		s = 0;
	else
		s = (short) ((8192L*bttv->hactive / field_width - 8192L + 1) 
				>> 1);

	VERBOSE("doScale, hscale=%d\n", s);
	r->e_hscale_hi = r->o_hscale_hi = (s >> 8) & 0xff;
	r->e_hscale_lo = r->o_hscale_lo = s & 0xff;

	if (bttv->field_height == (vactive >> 1) || bttv->field_height == -2)
		s = 0;
	else
		s = (short) ((0x10000L + 512L - 
					((512L*vactive /bttv->field_height + 1) 
					>> 1)) & 0x1FFFL);

	VERBOSE("doScale, vscale=%d\n", s);
	r->e_vscale_hi = r->o_vscale_hi = (s >> 8) & 0x1f;
	r->e_vscale_lo = r->o_vscale_lo = s & 0xff;
}

static void scalerInit(bttv_t *bttv)
{
	local_regs_t *r = bttv->regs;
	int f, a;

	f = (r->o_hscale_hi << 8) | r->o_hscale_lo;
	bttv->hdelay = ((f + 4096) * ((r->o_hdelay_msb << 8) | 
				r->o_hdelay_lo) + 2048) >> 12;
	bttv->hactive = ((f + 4096) * ((r->o_hactive_msb << 8) | 
				r->o_hactive_lo) + 2048) >> 12;
	f = (r->o_vscale_hi << 8) | r->o_vscale_lo;
	a = (r->o_vactive_msb << 8) | r->o_vactive_lo;
	bttv->field_height = (f == 0) ? a : (512L*a / (0x2200L - f) - 2);
	VERBOSE("hdelay=%d, hactive=%d, field_height=%d\n",
			bttv->hdelay, bttv->hactive, bttv->field_height);
	doScale(bttv);
}
/* end of code from ccd.c */

#define XT0FRQ 28636363

/* always assume pre devisor = 2, post divisor = 6 */
double bttv_set_pll(bttv_t *bttv, double freq) {
	local_regs_t *r = bttv->regs;
	double times = 12*65536*freq/XT0FRQ;
	int ipart;
/*	int mult = (int)round(times); */
	int mult = (int)times;
        if (times - (double) mult >= 0.5) mult++;
	r->pll_x = 1;
	r->pll_c = 0;
	r->pll_f_lo = mult&0xFF;
	r->pll_f_hi = (mult>>8)&0xFF;
	ipart = mult>>16;
	assert(ipart >= 6 && ipart < 64);
	r->pll_i = ipart;
	assert(times > 0);
#if 0
	DEBUG("setting PLL to %lf (err=%lf)\n", XT0FRQ*(double)mult/(12*65536),
			XT0FRQ*(double)mult/(12*65536) - freq);
#endif
	return XT0FRQ*(double)mult/(12*65536) - freq;
}

double bttv_get_pll(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	double mult = (r->pll_f_lo+256*r->pll_f_hi)/65536. + r->pll_i;
	double div = ((r->pll_c == 0) ? 6 : 4)*((r->pll_x == 0) ? 1 : 2);
	return XT0FRQ*mult/div;
}

void bttv_pll_up(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	int pll_f_hi = r->pll_f_hi;
	int pll_f_lo = r->pll_f_lo;
	int pll_i = r->pll_i;
	if (pll_f_lo == 255) {
		r->pll_f_lo = 0;
		if (pll_f_hi == 255) {
			r->pll_f_hi = 0;
			if (pll_i >= 64) assert(0);
			else r->pll_i = pll_i + 1;
		} else r->pll_f_hi = pll_f_hi + 1;
	} else r->pll_f_lo = pll_f_lo + 1;
}
		
void bttv_schedule_down(bttv_t *bttv) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->updown = -1;
	pthread_mutex_unlock(&bttv->cmutex);
}
		
void bttv_schedule_updown(bttv_t *bttv, int updown) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->updown = updown;
	pthread_mutex_unlock(&bttv->cmutex);
}

void bttv_schedule_pll(bttv_t *bttv, double newfrq) {
	pthread_mutex_lock(&bttv->cmutex);
	bttv->newfrq = newfrq;
	pthread_mutex_unlock(&bttv->cmutex);
}

void bttv_set_vtotal(bttv_t *bttv, int vtotal) {
	local_regs_t *r = bttv->regs;
	vtotal--;
	r->vtotal_hi = (vtotal>>8)&0xff;
	//DEBUG("setting vtotal to %d\n", vtotal + 1);
	r->vtotal_lo = vtotal&0xff;
}

void bttv_pll_down(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	int pll_f_hi = r->pll_f_hi;
	int pll_f_lo = r->pll_f_lo;
	int pll_i = r->pll_i;
	if (pll_f_lo == 0) {
		r->pll_f_lo = 255;
		if (pll_f_hi == 0) {
			r->pll_f_hi = 255;
			if (pll_i <= 6) assert(0);
			else r->pll_i = pll_i - 1;
		} else r->pll_f_hi = pll_f_hi - 1;
	} else r->pll_f_lo = pll_f_lo - 1;
}
		
int bttv_get_vtotal(bttv_t *bttv) {
	local_regs_t *r = bttv->regs;
	return (r->vtotal_lo|(r->vtotal_hi<<8)) + 1;
}

int bttv_init(bttv_t *bttv, char *dev, long chan) {
	int i;
	unsigned long int device, addr = 0, pcireg;
	struct video_audio audio;
	struct v4l2_control amute;
	local_regs_t *r;
	assert(bttv && dev);

	/* defaults */
	bttv->fd = -1;
	bttv->memfd = -1;
	bttv->map = (char*)-1;
	bttv->regs = (local_regs_t*)-1;

	if (geteuid() != 0) {
		ERROR("we need to access local registers of the Bt8x[89] chip; root privileges are needed\n");
		return -1;
	}

	if ((bttv->fd = open(dev, O_RDWR)) == -1) {
		ERROR("opening %s: %s\n", dev, strerror(errno));
		return -1;
	}
	VERBOSE("%s successfully openend\n", dev);

	memset(&bttv->vc,0,sizeof(bttv->vc));
	bttv->vc.type=VID_TYPE_CAPTURE;

	if (ioctl(bttv->fd, VIDIOCGCAP, &bttv->vc)) {
		ERROR("getting video capabilities, not v4l device?\n");
		close(bttv->fd);
		return -1;
	}

	/* aquire info on grab buffers... */
	if (ioctl(bttv->fd, VIDIOCGMBUF, &bttv->buf_info) == -1) {
		ERROR("VIDIOCGMBUF, bailing out\n");
		close(bttv->fd);
		return -1;
	}

	/* check number and size */
	if (bttv->buf_info.frames < 4) {
		ERROR("not enough frames in provided capture buffer\n");
		close(bttv->fd);
		return -1;
	}

	if (bttv->buf_info.size/bttv->buf_info.frames != 2129920) {
		ERROR("buffers have unexpected size, don't know how to do interlacing, bailing out\n");
		close(bttv->fd);
		return -1;
	}

	/* and map them */
	bttv->map = mmap(0, bttv->buf_info.size, PROT_READ|PROT_WRITE,
			MAP_SHARED, bttv->fd, 0);
	if ((int)bttv->map == -1) {
		ERROR("mapping %d capture buffers of size %d\n", 
				bttv->buf_info.frames, bttv->buf_info.size);
		return -1;
	} else VERBOSE("successfully mapped %d buffers of total size %d\n", 
			bttv->buf_info.frames, bttv->buf_info.size);
	
	for (i = 0; i < 4; i++) {
		bttv->maps[i].format = VIDEO_PALETTE_RAW;
		bttv->maps[i].frame = i;
		/* width and height seem to be ignored in raw capture more, 
		 * but values must be 'reasonable' (nonzero etc) */
		bttv->maps[i].width = 826;
		bttv->maps[i].height = 288;
	}

	/* try to find local io space of bttv chip */
	if (iopl(3) < 0) {
		ERROR("iopl(3): %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}
	
	i = 0;
	do {
		pcireg = 0x80000000 | (i << 8);

		outl(pcireg, 0xCF8);
		device = inl(0xCF8 + 4);

		switch (device) {
			case BT848:
			case BT849:
			case BT878:
			case BT879:
				bttv->type = device;
				outl(pcireg + 0x10, 0xCF8);
				addr =  inl(0xCF8 + 4)&0xfffffff0;
		}
	} while (++i < 0x10000 && addr == 0);

	if (addr == 0) {
		ERROR("unable to find BTTV chip, is your V4L device a BT8[47][89](A)?\n");
		bttv_uninit(bttv);
		return -1;
	} else VERBOSE("io space of chip is at 0x%lx\n", addr);

	switch (bttv->type) {
		case BT848:
		case BT849:
			ERROR("chip seems to be a BT84[89], unsupported at the moment\n");
			iopl(0);
			bttv_uninit(bttv);
			return -1;
			break;
		case BT878:
		case BT879:
			VERBOSE("chip seems to be a BT87%c\n", 
	    bttv->type == BT878 ? '8' : '9');
			break;
	}
	
	if (iopl(0) < 0) {
		ERROR("iopl(0): %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}

	/* map the registers */
	if ((bttv->memfd = open("/dev/mem", O_RDWR)) < 0) {
		ERROR("unable to open /dev/mem: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	}
	
	bttv->regs = mmap(0, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED,
			bttv->memfd, addr);

	if ((int)bttv->regs == -1) {
		ERROR("unable to mmap local registers: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	} else VERBOSE("successfully mapped local registers\n");
	
	r = bttv->regs;

	/* Save old values for restore at exit */

	bttv->old_regs = *r;
	bttv->old_vtotal = bttv_get_vtotal(bttv);
	bttv->old_pll= bttv_get_pll(bttv);

	/* setting up internal registers */

	scalerInit(bttv);
	r->xtsel = 1; 		/* select crystal 0, 28.636363MHz */
	bttv_set_pll(bttv, 35437500);
	r->tgcki = 1; 		/* enable pll */
	r->dec_fieldalign = 1;	/* align fields */
	r->dec_rate = 0;	/* don't drop frames */
	bttv_set_vtotal(bttv, 625);
	r->format = 3; 		/* 1=NTSC 3=PALBG, setting format */
	r->color_even = r->color_odd = 14; 
				/* setting Raw8X mode */ 
	r->test = 239;		/* don't try to sync */
	r->ext_frame = 1;	/* extend capture size */
	r->fifo_enable = 1;
	r->capture_even = r->capture_odd = 1;
	bttv->synced = -1; 
	bttv->queued0 = -1; 
	bttv->queued1 = -1; 
	bttv->exported = -1;
	bttv->avail = -1;
	pthread_mutex_init(&bttv->bmutex, NULL);
	pthread_mutex_init(&bttv->cmutex, NULL);
	pthread_cond_init(&bttv->acond, NULL);
	bttv->updown = 0;
	bttv->newfrq = 35437500;
	
  	/* unmute audio */
  
	memset(&audio,0,sizeof(audio)); 

	if (ioctl(bttv->fd,VIDIOCGAUDIO,&audio) == -1) {
    		ERROR("get audio caps failed: %s\n", strerror(errno));
    		bttv_uninit(bttv);
		exit(1);
	}

	audio.flags &= ~VIDEO_AUDIO_MUTE;

	if (ioctl(bttv->fd,VIDIOCSAUDIO,&audio)==-1) {
		ERROR("set audio caps failed: %s\n", strerror(errno));
		bttv_uninit(bttv);
		exit(1);;
	}

	/* turn off automute 
	 * Version 0.9.x of the bttv driver mutes the audio if it does not
	 * detect a valid signal. When the test register is set to 239 the
	 * driver considers the signal as not valid and mutes the audio. There
	 * are two possible solutions: load the driver with the option
	 * automute=0 or call the driver with the VIDIOC_S_CTRL ioctl and set
	 * automute to 0.  The second option may only work with version 0.9. I
	 * don't have version 0.7.x to test it.  VIDIOC_S_CTRL exists only for
	 * v4l2 but it is safe to use it to set automute to 0 since it only
	 * changes the value of the variable.  
	 */
       
	bttv->old_automute = 0;
	amute.id = V4L2_CID_PRIVATE_AUTOMUTE;
	if (ioctl(bttv->fd, VIDIOC_G_CTRL, &amute) == -1) 
		ERROR("Error getting automute state: %s\n", strerror(errno));
    	else {
		if (amute.value==1) {
			bttv->old_automute = 1;
			amute.value = 0;
			if (ioctl(bttv->fd, VIDIOC_S_CTRL, &amute) == -1)
	  			ERROR("Error setting automute state: %s\n", 
						strerror(errno));
		}
	}
    
	/* channel select, some of it copied from tvi_v4l in MPlayer,
	 * original code written and (C) by Alex Beregszaszi */
	bttv->channels = calloc(1, sizeof(*bttv->channels)*bttv->vc.channels);
	if (!bttv->channels) {
		ERROR("calloc failure: %s\n", strerror(errno));
		bttv_uninit(bttv);
		return -1;
	} 

	for (i = 0; i < bttv->vc.channels; i++) {
		struct video_channel *vch = bttv->channels + i;
		vch->channel = i;
		if (ioctl(bttv->fd, VIDIOCGCHAN, vch) == -1) {
			ERROR("get channel failed: %s\n", strerror(errno));
			bttv_uninit(bttv);
			return -1;
		} 
		
		VERBOSE(" %d: %s: %s%s%s%s (tuner:%d, norm:%d)\n", i,
				vch->name,
				(vch->flags&VIDEO_VC_TUNER) ?  "[tuner]" : "",
				(vch->flags&VIDEO_VC_AUDIO) ?  "[audio]" : "",
				(vch->flags&VIDEO_TYPE_TV) ?  "[tv]" : "",
				(vch->flags&VIDEO_TYPE_CAMERA) ? "[cmera]" : "",
				vch->tuners, vch->norm);
	}

	bttv_set_frequency(bttv,chan);

	return 0;
}

/* if chan == 0 -> dont mess with input settings
 * if chan == -1 -> select composite input (on my card, at least)
 * if chan > 0 -> set frequency */
void bttv_set_frequency(bttv_t *bttv,long chan)
{
        struct video_tuner tuner;

	if (chan == 0) return;

	if (ioctl(bttv->fd, VIDIOCSCHAN, 
				&bttv->channels[(chan < 0)?1:0]) == -1) {
		ERROR("set tv input failed... bad!\n");
		bttv_uninit(bttv);
		exit(1);
	}

	if (chan < 0) return;

	/* Assume (possibly incorrectly), that only one tuner exists,
	 * and that it is connected to channel 0 */
	tuner.tuner = 0;

	if (ioctl(bttv->fd, VIDIOCGTUNER, &tuner) == -1) {
		ERROR("error getting info from tuner: %s\n", strerror(errno));
		bttv_uninit(bttv);
                exit(1);
	}

	if (!(tuner.flags&VIDEO_TUNER_LOW)) chan /= 1000;

	if (ioctl(bttv->fd, VIDIOCSFREQ, &chan) == -1) {
		ERROR("set tuner freq failed: %s\n", strerror(errno));
		bttv_uninit(bttv);
                exit(1);
	}
}

static void queue(bttv_t *bttv, int i) {
	//DEBUG("queueing buffer %d for capture\n", i);
	if (ioctl(bttv->fd, VIDIOCMCAPTURE, &bttv->maps[i]) == -1) {
		ERROR("queueing buffer for capture\n");
	}
}

static void my_sync(bttv_t *bttv, int i) {
	//DEBUG("waiting for buffer %d to be full\n", i);
	do errno = 0;
	while (ioctl(bttv->fd, VIDIOCSYNC, &bttv->maps[i].frame) < 0 &&
			(errno == EAGAIN || errno == EINTR)); 
	if (errno != 0) ERROR("my_sync failed: %s\n", strerror(errno));
}

void bttv_capture_thread(void *data) {
	bttv_t *bttv = data;
	char *frame;
	int frameno = 0;
	int tmp;
	DEBUG("starting capture thread\n");
	
	pthread_mutex_lock(&bttv->bmutex);
	bttv->synced = -1;
	bttv->avail = 2;
	bttv->exported = 3;
	pthread_mutex_unlock(&bttv->bmutex);

	bttv->queued0 = 0;
	bttv->queued1 = 1;
	bttv->finfos[frameno%10].pll = bttv_get_pll(bttv);
	bttv->finfos[frameno%10].vtotal = bttv_get_vtotal(bttv);
	bttv->framenoes[bttv->queued0] = frameno++;
	queue(bttv, 0);
	queue(bttv, 1);
	while (1) {
		pthread_testcancel();
		my_sync(bttv, bttv->queued0);
		pthread_mutex_lock(&bttv->cmutex);
		bttv_set_vtotal(bttv, 625 + bttv->updown);
		bttv->finfos[frameno%10].vtotal = 625 + bttv->updown; 
		bttv->updown = 0;
		bttv->finfos[frameno%10].pll = 
			bttv->newfrq + bttv_set_pll(bttv, bttv->newfrq);
		pthread_mutex_unlock(&bttv->cmutex);
		pthread_mutex_lock(&bttv->bmutex);
		bttv->framenoes[bttv->queued1] = frameno++;
		if (bttv->avail >= 0 && bttv->synced < 0) {
			//VERBOSE("normal\n");
			bttv->synced = bttv->queued0;
			bttv->queued0 = bttv->queued1;
			bttv->queued1 = bttv->avail;
			bttv->avail = -1;
		} else if (bttv->avail < 0 && bttv->synced >= 0) {
			VERBOSE("input framedrop\n");
			tmp = bttv->synced;
			bttv->synced = bttv->queued0;
			bttv->queued0 = bttv->queued1;
			bttv->queued1 = tmp;
		} else assert(0);
		frame = bttv->map + bttv->buf_info.offsets[bttv->
			maps[bttv->synced].frame];
		pthread_cond_signal(&bttv->acond);
		pthread_mutex_unlock(&bttv->bmutex);
		queue(bttv, bttv->queued1);
	}
}

char *bttv_getnext(bttv_t *bttv, int *no) {
	char *out = NULL;
	pthread_mutex_lock(&bttv->bmutex);
	if (bttv->synced < 0) {
		/* no buffer currently available, so wait... */
		//printf("waiting for buffer\n");
	 	pthread_cond_wait(&bttv->acond, &bttv->bmutex);
	}
	bttv->avail = bttv->exported;
	bttv->exported = bttv->synced;
	bttv->synced = -1;
	*no = bttv->framenoes[bttv->exported];
	out = bttv->map + bttv->buf_info.offsets[bttv->
		maps[bttv->exported].frame];
	pthread_mutex_unlock(&bttv->bmutex);
	return out;
}
	
void bttv_uninit(bttv_t *bttv) {
	int i;
	local_regs_t *r;
	struct video_audio audio;
	struct v4l2_control amute;

	/* mute audio */
	ioctl(bttv->fd,VIDIOCGAUDIO,&audio);
	audio.flags|=VIDEO_AUDIO_MUTE;
	ioctl(bttv->fd,VIDIOCSAUDIO,&audio);

	/* restore automute if it was set */

	if (bttv->old_automute == 1) {
		amute.id=V4L2_CID_PRIVATE_AUTOMUTE;
		amute.value=1;
		if (ioctl(bttv->fd,VIDIOC_S_CTRL,&amute)==-1)
			ERROR("Error resetting automute state: %s\n", 
					strerror(errno));
	}
	free(bttv->channels);

	if ((int)bttv->regs != -1) {
		/* restore the initial values of the BT848/BT878 registers */
		r=bttv->regs;

		r->xtsel = bttv->old_regs.xtsel;
		bttv_set_pll(bttv, bttv->old_pll);
		r->tgcki = bttv->old_regs.tgcki;
		r->dec_fieldalign = bttv->old_regs.dec_fieldalign;
		r->dec_rate = bttv->old_regs.dec_rate;
		bttv_set_vtotal(bttv, bttv->old_vtotal);
		r->format = bttv->old_regs.format;
		r->color_even =bttv->old_regs.color_even; 
		r->color_odd = bttv->old_regs.color_odd; 
		r->test = 1;
		r->ext_frame = bttv->old_regs.ext_frame;
		r->fifo_enable = bttv->old_regs.fifo_enable;
		r->capture_even = bttv->old_regs.capture_even;
		r->capture_odd = bttv->old_regs.capture_odd;

		munmap((void *)(bttv->regs), 0x1000);
		VERBOSE("unmapped local io regs\n");
	}

	if (bttv->memfd != -1) {
		close(bttv->memfd);
		VERBOSE("closed /dev/mem\n");
	}

	for (i = 0; i < bttv->buf_info.frames; i++) ioctl(VIDIOCSYNC, i);

	if ((int)bttv->map != -1) {
		munmap(bttv->map, bttv->buf_info.size);
		VERBOSE("unmapped capture buffers\n");
	}

	if (bttv->fd != -1) {
		close(bttv->fd);
		VERBOSE("closed video device\n");
	}
	VERBOSE("Exiting cc2y4m...\n");
}

