/* cc2y4m.c - main module
 *
 * Sighandler (C) 2004 by RB
 * The rest Copyright (C) 2003 Rik Snel <snel@phys.uu.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; only version 2 of the License, not
 * any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: cc2y4m.c 64 2004-08-01 10:34:42Z rsnel $
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include "cc2y4m.h"
#include "bttv.h"
#include "lock.h"
#include "esnelmDecode.h"
#include "mplayerKeys.h"
#include "parser.h"
#include <SDL/SDL.h>

#include <signal.h>

const struct lock_poldetect lock_defaults[] = { {
		.polx=1975,
		.poly=50,
		.poltresh=16,
 		.polnorm=0,
		.bright_adjust=135,
		.center=180,
		.filter=0,
		.refpoint=refpoint_m1,
		.locktresh=50,
	}, {
		.polx=145,
		.poly=50,
		.poltresh=45,
		.polnorm=0,
		.bright_adjust=200,
		.center=136,
		.filter=1,
		.refpoint=refpoint_m2,
		.locktresh=40,
	}
};

const unsigned int lock_pol_methods = 
	sizeof(lock_defaults)/sizeof(struct lock_poldetect);

/* Default method for locking and polarity detection, Modify with option -m */
const struct lock_poldetect *method = &lock_defaults[0];

static bttv_t bttv;
static pthread_t capture;
static pthread_t lockthread;
static pthread_attr_t attr;

extern int brightness;
extern int contrast;
extern int saturation;

int readcfg = 1;
int mplayerkeys = 0;
int channel_no = -1;

void sighandler(int s) {
	if(s ==	SIGPIPE)
		VERBOSE("SIGPIPE received. Terminating...\n");
	if(s == SIGTERM)
		VERBOSE("SIGTERM received. Terminating...\n");
	if(s == SIGINT)
		VERBOSE("SIGINT received. Terminating...\n");

	VERBOSE("Waiting for capture thread...\n");
	pthread_cancel(capture);
	pthread_join(capture, NULL);

	VERBOSE("Waiting for lock thread...\n");
	pthread_cancel(lockthread);
	pthread_join(lockthread, NULL);

	pthread_attr_destroy(&attr);
	VERBOSE("Resetting bttv...\n");
	bttv_uninit(&bttv);
	exit(0);
}

static void yuv4mpeg2_out(unsigned char* yuv, int fields, FILE *fp) {
	int i;
	fprintf(fp, "FRAME\n");
	fwrite(yuv, 1, 384*288*fields, fp);
	for (i = 0; i < 144; i++) {
		fwrite(yuv+ 384*(288+i)*fields, 1, 192, fp);
		if (fields == 2) fwrite(yuv + 384*(576+2*i) + 192, 1, 192, fp);
	}
	for (i = 0; i < 144; i++) {
		fwrite(yuv+ 384*(432+i)*fields, 1, 192, fp);
		if (fields == 2) fwrite(yuv + 384*(864+2*i) + 192, 1, 192, fp);
	}
}

extern char *optarg;
extern int optind, opterr, optopt;

int quiet = 0, debug = 0;

/*  Filter (By Pinguim) v.2
    
    On all TV-channels I have checked there is a black area on the left-side of
    the visual image. The flickering of about 10Hz (as addressed in the FAQ)
    can be clearly seen in this area. 
 
    First for every visible line the average black-level is calculated. The 
    average is taken, because there is a blocked pattern, which could ruin the 
    average if we take just one value.
    Now every line in the field is adjusted so, that the earlier calculated
    average black-level of the line is adjusted to the average black-level of
    the previous field.
    
    Second the average blacklevel of the field is calculated by averaging all
    the mean levels of the visible lines. The int fieldmean holds the average
    value for the previous field. It is this value that is used to correct
    the current field. one should use the value of the current field, but
    the level does not seem to change very fast. This way there is no
    overhead introduced for calculating the meanvalue of the field first.
    
    This reduced the 10Hz interfering signal significantly. 
    
    What if it doesn't work for you?
    Maybe you can adjust it to work for you. The main thing this filter requires
    is a vertical area (ussually black, but any colour would do). The horizontal
    position of the area is marked by FILT_X_MIN and FILT_X_MAX. The vertical
    range of the area is defined by the vertical area of the visible image 
    (there is no sense in filtering non-visible area), which can be set by
    VIS_Y_MIN and VIS_Y_MAX. The horizontal range that will be filtered is
    defined by VIS_X_MIN and VIS_X_MAX.
*/
#define VID_RAW_HEIGHT 288
#define VID_RAW_WIDTH 2268

#define FILT_X_MIN 0
#define FILT_X_MAX 50
#define VIS_X_MIN 150
#define VIS_X_MAX 2268
#define VIS_Y_MIN 0
#define VIS_Y_MAX VID_RAW_HEIGHT

static int fieldmean;
static void filter(char *in, int offset) {
	int y, x, linemean, newfieldmean = 0;
	if (offset < 544320 && offset > -30) {
		for (y = VIS_Y_MIN; y < VIS_Y_MAX; y++) {
			linemean = 0;
			for (x = y*VID_RAW_WIDTH + FILT_X_MIN; x < y*VID_RAW_WIDTH + FILT_X_MAX; x++)
				linemean += (*(in + offset + x));
			linemean /= FILT_X_MAX - FILT_X_MIN;
			newfieldmean += linemean;
			for (x = y*VID_RAW_WIDTH + VIS_X_MIN; x < y*VID_RAW_WIDTH + VIS_X_MAX; x++)
				(*(in + offset + x)) = (*(in + offset + x)) + fieldmean - linemean;
		}
		fieldmean = newfieldmean/(VIS_Y_MAX - VIS_Y_MIN);
	}
}

static void poldetect(char *in, int offset) {
	if (offset < 544320 && offset > -30) {
		if (*(in + offset + method->poly*2268 + method->polx) 
				> method->poltresh) {
			setChroma(1-method->polnorm);
			setPolarity(1-method->polnorm);
		} else {
			setChroma(method->polnorm);
			setPolarity(method->polnorm);
		}
	} else {
		setChroma(0);
		setPolarity(0);
	}
}

int main(int argc, char **argv) {
	SDL_Color colors[256];
	SDL_Surface *screen;
	char *out;
	char *home;
	int j;
	unsigned char yuv[4*384*288];
	lock_t lock;
	int i, k, opt, offset = 0, fields = 1;
	long chan = 0;
	char *in, *field, *dev = "/dev/video0", *key,conf_file[100];
	struct channel_s *chanlist;

	while ((opt = getopt(argc, argv, "c:Vn:hkDqd:im:B:C:S:")) != -1) {
		if (opt == '?' || opt == ':') exit(1);
		switch (opt) {
			case 'V':
				printf("cc2y4m version is " VERSION ", (C) 2004 Rik Snel et al. (see AUTHORS for details)\nPAL color decoding (C) 1999 Ewald Snel.\nThis is free software; see the source for copying conditions. The is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
				exit(1);
			case 'h':
				printf("Example: cc2y4m -i | mplayer - -zoom\n");
				exit(1);
			case 'D':
				debug = 1;
				break;
			case 'k':
				mplayerkeys = 1;
				break;
			case 'q':
				quiet = 1;
				break;
			case 'd':
				dev = optarg;
				break;
			case 'n':
				readcfg = 0;
				break;
			case 'i':
				fields = 2;
				break;
		        case 'm':
				i = atoi(optarg);
				if (i < 1 || i > lock_pol_methods) {
					ERROR("The method must be between 1 and %d\n",lock_pol_methods);
					exit(1);
			  	}
				method=&lock_defaults[i-1];
				break;
			case 'B':
				brightness = atoi(optarg);
				break;
			case 'C':
				contrast = atoi(optarg);
				break;
			case 'S':
				saturation = atoi(optarg);
				break;
			case 'c':
				channel_no = atoi(optarg);
				break;
		}

	}
	if (optind != argc) {
		ERROR("unrecognized stuff on the command line, please fix\n");
		exit(1);
	}

	signal(SIGPIPE, sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGINT, sighandler);

	home = getenv("HOME");
	if (!home || !readcfg) {
                VERBOSE("$HOME not set or ~/.cc2y4m disabled. Using defaults.\n");
	} else {
		int len;
		len = strlen(home) + 1 + strlen(".cc2y4m") + 1;
		if (len > 100) {
			VERBOSE("path in $HOME is too long, unable to read ~/.cc2y4m, using defaults");
		} else {
			conf_file[0]='\0';
			strcat(conf_file,home);
			strcat(conf_file,"/"); strcat(conf_file,".cc2y4m");
	
			if((chanlist=parse_file(conf_file))==NULL) {
				VERBOSE("Could not open ~/.cc2y4m. Using defaults.\n");
			}
       		}
	}


	if (bttv_init(&bttv, dev, chan) == -1) {
		exit(1);
	}

	if(chanlist) {
		while (channel_no > 0) {
			channel_no--;
			chanlist = chanlist->next;
			if (!chanlist) {
				ERROR("illegal channel number specified with -c");
				exit(1);
			}
		}
		method=&(chanlist->p);
		bttv_set_frequency(&bttv,chanlist->freq);
	}

	lock_init(&lock, &bttv);

	if (debug) {
		if (SDL_Init(SDL_INIT_VIDEO) < 0) {
			ERROR("Unable to init SDL: %s\n", SDL_GetError());
			exit(1);
		}
		atexit(SDL_Quit);

		for (i = 0; i < 256; i++) {
			colors[i].r = i;
			colors[i].g = i;
			colors[i].b = i;
		}

		colors[255].r = 255;
		colors[255].g = 127;
		colors[255].b = 127;
		colors[254].r = 127;
		colors[254].g = 255;
		colors[254].b = 127;

		screen = SDL_SetVideoMode(300, 200, 8, SDL_HWSURFACE);
		if (!screen) {
			ERROR("%s\n", SDL_GetError());
			exit(1);
		}

		if (SDL_MUSTLOCK(screen)) {
			ERROR("SDL screen must be locked, but locking not implemented (lazy)\n");
			exit(1);
		}

		SDL_SetColors(screen, colors, 0, 256);

		out = (Uint8*)screen->pixels;
	}

	InitializeDecoderMMX();

	pthread_attr_init(&attr);
	//pthread_attr_setschedpolicy(&attr, SCHED_RR);
	if (pthread_create(&capture, &attr, 
				(void*)&bttv_capture_thread, &bttv)) {
		ERROR("starting capture thread: %s\n", strerror(errno));
		bttv_uninit(&bttv);
		exit(1);
	}

	if (pthread_create(&lockthread, &attr, (void*)lock_thread, &lock)) {
		ERROR("starting locking thread: %s\n", strerror(errno));
		bttv_uninit(&bttv);
		exit(1);
	}
	
	for (i = 0; i < 384*288*fields; i++) yuv[i] = 0;
	for (i = 0; i < 384*288*fields; i++) 
		yuv[i+384*288*fields] = 128;

	printf("YUV4MPEG2 W384 H%d F25000000:1000000 I%c A%d:1\n",
			fields == 2 ? 576 : 288, fields == 2 ? 't' : 'p',
			fields);

        /* Setup to catch keypresses on MPlayer */
        if (mplayerkeys) { 
		init_mplayer_keys();
		if(chanlist) mplayer_set_title(chanlist->name);
	}

	while (1) {
                /* Test for a keypress */
        	if (mplayerkeys) { 

                if(chanlist && (key=get_mplayer_keys())!=NULL) {
                        if(!strncmp(key,"Up",2)) {
                                chanlist=chanlist->next;
                                bttv_set_frequency(&bttv,chanlist->freq);
                                mplayer_set_title(chanlist->name);
				method=&(chanlist->p);
                                lock.state=0;
                        }
                        else if(!strncmp(key,"Down",3)) {
                                chanlist=chanlist->prev;
                                bttv_set_frequency(&bttv,chanlist->freq);
                                mplayer_set_title(chanlist->name);
				method=&(chanlist->p);
                                lock.state=0;
                        }
                }

		} 

		/* input */
		in = lock_getnext(&lock, &offset);

		if (debug) {
			if (offset < 544320 && offset > -30) {
				for (i = 0; i < 300; i++) {
					for (j = 0; j < 100; j++) {
						*(out + 2*j*screen->pitch + i) =
							*(in + offset + j*2268 +
								method->polx-275
									+ i)*2;
						*(out + (2*j+1)*screen->pitch 
							+ i) = *(in + 1064960 
							+ offset + j*2268 + 
							method->polx-275 + i)*2;
					}
				}
				*(out + 2*method->poly*screen->pitch + 
						275) = 255;
				SDL_UpdateRect(screen, 0, 0, 0, 0);
			}
		}

		for (k = 0; k < fields; k++) {
			field = in + k*1064960;

			if (method->filter) filter(field, offset);
			poldetect(field, offset);

			/* enable software corrections for small drift */
			if (offset < -20 || offset > 20) offset = 0;
			decodeFrameEsnelColorMMX(yuv + k*384, 
					yuv + 384*288*fields + k*192,  
					yuv + 384*432*fields + k*192,  
					field+4*2268, offset+20, fields*384);
		}

		/* output */
		yuv4mpeg2_out(yuv, fields, stdout);
	}

	pthread_join(capture, NULL);
	pthread_join(lockthread, NULL);
	pthread_attr_destroy(&attr);

	bttv_uninit(&bttv);
	exit(0);
}
